import * as jwt from "jsonwebtoken";
import { Body, Get, JsonController, OnUndefined, Param, Post } from 'routing-controllers';
import { OpenAPI, ResponseSchema } from 'routing-controllers-openapi';
import { getConnectionManager, Repository } from 'typeorm';
import { config } from '../config';
import { InvalidCredentialsError } from '../errors/AuthError';
import { RunnerEmailNeededError, RunnerNotFoundError } from '../errors/RunnerErrors';
import { RunnerGroupNotFoundError } from '../errors/RunnerGroupErrors';
import { RunnerOrganisationNotFoundError } from '../errors/RunnerOrganisationErrors';
import { CreateSelfServiceCitizenRunner } from '../models/actions/create/CreateSelfServiceCitizenRunner';
import { CreateSelfServiceRunner } from '../models/actions/create/CreateSelfServiceRunner';
import { Runner } from '../models/entities/Runner';
import { RunnerGroup } from '../models/entities/RunnerGroup';
import { RunnerOrganisation } from '../models/entities/RunnerOrganisation';
import { ResponseSelfServiceRunner } from '../models/responses/ResponseSelfServiceRunner';


@JsonController('/runners')
export class RunnerSelfServiceController {
	private runnerRepository: Repository<Runner>;
	private orgRepository: Repository<RunnerOrganisation>;

	/**
	 * Gets the repository of this controller's model/entity.
	 */
	constructor() {
		this.runnerRepository = getConnectionManager().get().getRepository(Runner);
		this.orgRepository = getConnectionManager().get().getRepository(RunnerOrganisation);
	}

	@Get('/me/:jwt')
	@ResponseSchema(ResponseSelfServiceRunner)
	@ResponseSchema(RunnerNotFoundError, { statusCode: 404 })
	@OnUndefined(RunnerNotFoundError)
	@OpenAPI({ description: 'Lists all information about yourself. <br> Please provide your runner jwt(that code we gave you during registration) for auth. <br> If you lost your jwt/personalized link please contact support.' })
	async get(@Param('jwt') token: string) {
		return (new ResponseSelfServiceRunner(await this.getRunner(token)));
	}

	@Post('/register')
	@ResponseSchema(ResponseSelfServiceRunner)
	@ResponseSchema(RunnerEmailNeededError, { statusCode: 406 })
	@OpenAPI({ description: 'Create a new selfservice runner in the citizen org. <br> This endpoint shoud be used to allow "everyday citizen" to register themselves. <br> You have to provide a mail address, b/c the future we\'ll implement email verification.' })
	async registerRunner(@Body({ validate: true }) createRunner: CreateSelfServiceCitizenRunner) {
		let runner = await createRunner.toEntity();

		runner = await this.runnerRepository.save(runner);
		return new ResponseSelfServiceRunner(await this.runnerRepository.findOne(runner, { relations: ['scans', 'group', 'scans.track', 'cards', 'distanceDonations', 'distanceDonations.donor', 'distanceDonations.runner', 'distanceDonations.runner.scans', 'distanceDonations.runner.scans.track'] }));
	}

	@Post('/register/:token')
	@ResponseSchema(ResponseSelfServiceRunner)
	@ResponseSchema(RunnerGroupNotFoundError, { statusCode: 404 })
	@OpenAPI({ description: 'Create a new selfservice runner in a provided org. <br> The orgs get provided and authorized via api tokens that can be optained via the /organisations endpoint.' })
	async registerOrganisationRunner(@Param('token') token: string, @Body({ validate: true }) createRunner: CreateSelfServiceRunner) {
		const org = await this.getOrgansisation(token);

		let runner = await createRunner.toEntity(org);
		runner = await this.runnerRepository.save(runner);

		return new ResponseSelfServiceRunner(await this.runnerRepository.findOne(runner, { relations: ['scans', 'group', 'scans.track', 'cards', 'distanceDonations', 'distanceDonations.donor', 'distanceDonations.runner', 'distanceDonations.runner.scans', 'distanceDonations.runner.scans.track'] }));
	}

	/**
	 * Get's a runner by a provided jwt token.
	 * @param token The runner jwt provided by the runner to identitfy themselves.
	 */
	private async getRunner(token: string): Promise<Runner> {
		let jwtPayload = undefined
		try {
			jwtPayload = <any>jwt.verify(token, config.jwt_secret);
		} catch (error) {
			throw new InvalidCredentialsError();
		}

		const runner = await this.runnerRepository.findOne({ id: jwtPayload["id"] }, { relations: ['scans', 'group', 'scans.track', 'cards', 'distanceDonations', 'distanceDonations.donor', 'distanceDonations.runner', 'distanceDonations.runner.scans', 'distanceDonations.runner.scans.track'] });
		if (!runner) { throw new RunnerNotFoundError() }
		return runner;
	}

	/**
	 * Get's a runner org by a provided registration api key.
	 * @param token The organisation's registration api token.
	 */
	private async getOrgansisation(token: string): Promise<RunnerGroup> {
		token = Buffer.from(token, 'base64').toString('utf8');

		const organisation = await this.orgRepository.findOne({ key: token });
		if (!organisation) { throw new RunnerOrganisationNotFoundError; }

		return organisation;
	}
}